/*	basic.c:	Basic movement functions for
 *			MicroEMACS
 *			(C)Copyright 1993 by Daniel Lawrence
 *
 * The routines in this file move the cursor around on the screen. They
 * compute a new value for the cursor, then adjust ".". The display code
 * always updates the cursor location, so only moves between lines, or
 * functions that adjust the top line in the window and invalidate the
 * framing, are hard.
 */
#include	<stdio.h>
#include	"estruct.h"
#include	"eproto.h"
#include	"edef.h"
#include	"elang.h"

/*
 * Move the cursor to the
 * beginning of the current line.
 * Trivial.
 */
int gotobol(f, n)
     int f,
         n;			/* argument flag and num */
{
#if UNDOCMD
    if(!undo_in_progress)
        undo_save_dot(f, n);
#endif
    curwp->w_doto = 0;
    return (TRUE);
}

/*
 * Move the cursor backwards by "n" characters. If "n" is less than zero call
 * "forwchar" to actually do the move. Otherwise compute the new cursor
 * location. Error if you try and move out of the buffer. Set the flag if the
 * line pointer for dot changes.
 */
int backchar(f, n)
     int f,
         n;			/* prefix flag and argument */
{
    register LINE *lp;

    if (n < 0)
	return (forwchar(f, -n));
#if UNDOCMD
    if(!undo_in_progress)
        undo_save_fn(TRUE, n, forwchar);
#endif
    while (n--)
    {
	if (curwp->w_doto == 0)
	{
	    if ((lp = lback(curwp->w_dotp)) == curbp->b_linep)
		return (FALSE);
	    curwp->w_dotp = lp;
	    curwp->w_doto = lused(lp);
	    curwp->w_flag |= WFMOVE;
	}
	else
	    curwp->w_doto--;
    }
#if	DBCS
    return (stopback());
#else
    return (TRUE);
#endif
}

/*
 * Move the cursor to the end of the current line. Trivial. No errors.
 */
int gotoeol(f, n)
     int f,
         n;			/* argument flag and num */
{
#if UNDOCMD
    if(!undo_in_progress)
        undo_save_dot(f, n);
#endif
    curwp->w_doto = lused(curwp->w_dotp);
    return (TRUE);
}

/*
 * Move the cursor forwards by "n" characters. If "n" is less than zero call
 * "backchar" to actually do the move. Otherwise compute the new cursor
 * location, and move ".". Error if you try and move off the end of the
 * buffer. Set the flag if the line pointer for dot changes.
 */
int forwchar(f, n)
     int f,
         n;			/* prefix flag and argument */
{
    if (n < 0)
	return (backchar(f, -n));
#if UNDOCMD
    if(!undo_in_progress)
        undo_save_fn(TRUE, n, backchar);
#endif
    while (n--)
    {
	if (curwp->w_doto == lused(curwp->w_dotp))
	{
	    if (curwp->w_dotp == curbp->b_linep)
		return (FALSE);
	    curwp->w_dotp = lforw(curwp->w_dotp);
	    curwp->w_doto = 0;
	    curwp->w_flag |= WFMOVE;
	}
	else
	    curwp->w_doto++;
    }
#if	DBCS
    return (stopforw());
#else
    return (TRUE);
#endif
}

/*
 * Move to a particular line.  argument (n) must be a positive integer for
 *	this to actually do anything
 */
int gotoline(f, n)
     int f,
         n;			/* prefix flag and argument */
{
    register int status;	/* status return */
    char arg[NSTRING];		/* buffer to hold argument */

/* get an argument if one doesnt exist */
    if (f == FALSE)
    {
	if ((status = mlreply(TEXT7, arg, NSTRING)) != TRUE)
	{
	/*                                    "Line to GOTO: " */
	    mlwrite(TEXT8);
	/*                              "[Aborted]" */
	    return (status);
	}
	n = asc_int(arg);
    }

    if (n < 1)			/* if a bogus argument...then leave */
	return (FALSE);

#if UNDOCMD
	if(!undo_in_progress)
		undo_save_dot(f, n);
	/* first, we go to the start of the buffer */
	curwp->w_dotp  = lforw(curbp->b_linep);
	curwp->w_doto  = 0;
	undo_in_progress = TRUE;
	status = forwline(f, n-1);
	undo_in_progress = FALSE;
	return(status);
#else
	/* first, we go to the start of the buffer */
        curwp->w_dotp  = lforw(curbp->b_linep);
        curwp->w_doto  = 0;
	return(forwline(f, n-1));
#endif
}

/*
 * Goto the beginning of the buffer. Massive adjustment of dot. This is
 * considered to be hard motion; it really isn't if the original value of dot
 * is the same as the new value of dot. Normally bound to "M-<".
 */
int gotobob(f, n)
     int f,
         n;			/* argument flag and num */
{
#if UNDOCMD
    if(!undo_in_progress)
        undo_save_dot(f, n);
#endif
    curwp->w_dotp = lforw(curbp->b_linep);
    curwp->w_doto = 0;
    curwp->w_flag |= WFMOVE;
    return (TRUE);
}

/*
 * Move to the end of the buffer. Dot is always put at the end of the file
 * (ZJ). The standard screen code does most of the hard parts of update.
 * Bound to "M->".
 */
int gotoeob(f, n)
     int f,
         n;			/* argument flag and num */
{
#if UNDOCMD
    if(!undo_in_progress)
        undo_save_dot(f, n);
#endif
    curwp->w_dotp = curbp->b_linep;
    curwp->w_doto = 0;
    curwp->w_flag |= WFMOVE;
    return (TRUE);
}

/*
 * Move forward by full lines. If the number of lines to move is less than
 * zero, call the backward line function to actually do it. The last command
 * controls how the goal column is set. Bound to "C-N". No errors are
 * possible.
 */
int forwline(f, n)
     int f,
         n;			/* argument flag and num */
{
    register LINE *dlp;

    if (n < 0)
	return (backline(f, -n));

/* if we are on the last line as we start....fail the command */
    if (curwp->w_dotp == curbp->b_linep)
	return (FALSE);

#if UNDOCMD
    if (!undo_in_progress)
        undo_save_fn(TRUE, n, backline);
#endif

/* if the last command was not note a line move,
	   reset the goal column */
    if ((lastflag & CFCPCN) == 0)
	curgoal = getccol(FALSE);

/* flag this command as a line move */
    thisflag |= CFCPCN;

/* and move the point down */
    dlp = curwp->w_dotp;
    while (n-- && dlp != curbp->b_linep)
	dlp = lforw(dlp);

/* reseting the current position */
    curwp->w_dotp = dlp;
    curwp->w_doto = getgoal(dlp);
    curwp->w_flag |= WFMOVE;
#if	DBCS
    return (stopback());
#else
    return (TRUE);
#endif
}

/*
 * This function is like "forwline", but goes backwards. The scheme is exactly
 * the same. Check for arguments that are less than zero and call your
 * alternate. Figure out the new line and call "movedot" to perform the
 * motion. No errors are possible. Bound to "C-P".
 */
int backline(f, n)
     int f,
         n;			/* argument flag and num */
{
    register LINE *dlp;

    if (n < 0)
	return (forwline(f, -n));


/* if we are on the last line as we start....fail the command */
    if (lback(curwp->w_dotp) == curbp->b_linep)
	return (FALSE);

#if UNDOCMD
    if(!undo_in_progress)
        undo_save_fn(TRUE, n, forwline);
#endif

/* if the last command was not note a line move,
	   reset the goal column */
    if ((lastflag & CFCPCN) == 0)
	curgoal = getccol(FALSE);

/* flag this command as a line move */
    thisflag |= CFCPCN;

/* and move the point up */
    dlp = curwp->w_dotp;
    while (n-- && lback(dlp) != curbp->b_linep)
	dlp = lback(dlp);

/* reseting the current position */
    curwp->w_dotp = dlp;
    curwp->w_doto = getgoal(dlp);
    curwp->w_flag |= WFMOVE;
#if	DBCS
    return (stopback());
#else
    return (TRUE);
#endif
}

/* Go back to the beginning of the current paragraph
 * here we look for a blank line or a character from
 * $paralead to delimit the beginning of a paragraph or
 * $fmtlead to delimit a line before the paragraph
 */
int gotobop(f, n)
     int f,
         n;			/* default Flag & Numeric argument */
{
    register int suc;		/* success of last backchar */
    register int c;		/* current character in scan */
    register char *sp;		/* ptr into character leadin lists */

    if (n < 0)			/* the other way...*/
	return (gotoeop(f, -n));

#if UNDOCMD
    if(!undo_in_progress)
    {
        undo_save_dot(f, n);
        undo_in_progress = TRUE;
    }
#endif

    while (n-- > 0)
    {				/* for each one asked for */

    /* first scan back until we are in a word */
	suc = backchar(FALSE, 1);
	while (!inword() && suc)
	    suc = backchar(FALSE, 1);
	curwp->w_doto = 0;	/* and go to the B-O-Line */

    /* scan back through the text */
	while (lback(curwp->w_dotp) != curbp->b_linep)
	{

	/* at blank line */
	    if (lused(curwp->w_dotp) == 0)
		break;

	/* current line start with member of $paralead? */
	    c = lgetc(curwp->w_dotp, 0);
	    sp = paralead;
	    while (*sp)
	    {
		if (c == *sp)
		    break;
		++sp;
	    }
	    if (c == *sp)
		break;

	/* last line start with member of $fmtlead? */
	    c = lgetc(lback(curwp->w_dotp), 0);
	    sp = fmtlead;
	    while (*sp)
	    {
		if (c == *sp)
		    break;
		++sp;
	    }
	    if (c == *sp)
		break;

	/* back one line... */
	    curwp->w_dotp = lback(curwp->w_dotp);
	}

    /* and then forward until we are in a word */
    /*		suc = forwchar(FALSE, 1); */
	suc = TRUE;
	while (suc && !inword())
	    suc = forwchar(FALSE, 1);
    }
    curwp->w_flag |= WFMOVE;	/* force screen update */

#if UNDOCMD
    if(!undo_in_progress)
        undo_in_progress = FALSE;
#endif
    return (TRUE);
}

/*
 * Go forword to the end of the current paragraph
 * looking for a member of $paralead or $fmtlead
 * or a blank line to delimit the start of the
 * next paragraph
 */
int gotoeop(f, n)
     int f,
         n;			/* default Flag & Numeric argument */

{
    register int suc;		/* success of last backchar */
    register int c;		/* current character in scan */
    register char *sp;		/* ptr into character leadin lists */

    if (n < 0)			/* the other way...*/
	return (gotobop(f, -n));
#if UNDOCMD
    if(!undo_in_progress)
    {
        undo_save_dot(f, n);
        undo_in_progress = TRUE;
    }
#endif

    while (n-- > 0)
    {				/* for each one asked for */

    /* first scan forward until we are in a word */
	suc = forwchar(FALSE, 1);
	while (!inword() && suc)
	    suc = forwchar(FALSE, 1);

    /* and go to the B-O-Line */
	curwp->w_doto = 0;

    /* of next line if not at EOF */
	if (suc)
	    curwp->w_dotp = lforw(curwp->w_dotp);

    /* scan forward */
	while (curwp->w_dotp != curbp->b_linep)
	{

	/* at blank line */
	    if (lused(curwp->w_dotp) == 0)
		break;

	/* current line start with member of $paralead? */
	    c = lgetc(curwp->w_dotp, 0);
	    sp = paralead;
	    while (*sp)
	    {
		if (c == *sp)
		    break;
		++sp;
	    }
	    if (c == *sp)
		break;

	/* current line start with member of $fmtlead? */
	    c = lgetc(curwp->w_dotp, 0);
	    sp = fmtlead;
	    while (*sp)
	    {
		if (c == *sp)
		    break;
		++sp;
	    }
	    if (c == *sp)
		break;

	/* forward one line... */
	    curwp->w_dotp = lforw(curwp->w_dotp);
	}

    /* and then backward until we are in a word */
	suc = backchar(FALSE, 1);
	while (suc && !inword())
	{
	    suc = backchar(FALSE, 1);
	}
	curwp->w_doto = lused(curwp->w_dotp);	/* and to the EOL */
    }
    curwp->w_flag |= WFMOVE;	/* force screen update */

#if UNDOCMD
    if(!undo_in_progress)
        undo_in_progress = FALSE;
#endif
    return (TRUE);
}

/*
 * This routine, given a pointer to a LINE, and the current cursor goal
 * column, return the best choice for the offset. The offset is returned.
 * Used by "C-N" and "C-P".
 */
int getgoal(dlp)
     register LINE *dlp;
{
    register int c;
    register int col;
    register int newcol;
    register int dbo;

    col = 0;
    dbo = 0;
    while (dbo != lused(dlp))
    {
	c = lgetc(dlp, dbo);
	newcol = col;
	if (c == '\t')
	    newcol += -(newcol % tabsize) + (tabsize - 1);
	else if (c < 0x20 || c == 0x7F)
	    ++newcol;
	++newcol;
	if (newcol > curgoal)
	    break;
	col = newcol;
	++dbo;
    }
    return (dbo);
}

/*
 * Scroll forward by a specified number of lines, or by a full page if no
 * argument. Bound to "C-V". The overlap in the arithmetic on the window size
 * is overlap between screens. This defaults to overlap value in ITS EMACS.
 *  Because this zaps the top line in the window, we have to do a hard update.
 */
forwpage(f, n)
     int f,
         n;			/* prefix flag and argument */
{
    register LINE *lp;

#if	BRIEF
	/*
 	 *	Brief allows the cursor after the end of a line.  It also
	 *	leaves the cursor on the same screen line on PgUp/PgDn
	 */
		int	line_offset = 0;

		/* find out how many lines into the screen is the dot */
        lp = curwp->w_linep;
		while(lp != curbp->b_linep && lp != curwp->w_dotp)
		{
			line_offset++;
			lp = lforw(lp);
		}
#endif

    if (f == FALSE)
    {
	n = curwp->w_ntrows - overlap;	/* Default scroll.      */
	if (n <= 0)		/* Forget the overlap   */
	    n = 1;		/* if tiny window.      */
    }
    else if (n < 0)
	return (backpage(f, -n));
    lp = curwp->w_linep;
#if UNDOCMD
    if(!undo_in_progress)
        undo_save_dot(f, n);
#endif
    while (n-- && lp != curbp->b_linep)
	lp = lforw(lp);
    curwp->w_linep = lp;
#if	BRIEF
    while (lp!=curbp->b_linep)
    {
	if(line_offset == 0)
            break;
	else
	{
	    line_offset--;
	    lp = lforw(lp);
	}
    }
    curwp->w_dotp  = lp;
    if (curwp->w_doto > lused(lp))
        curwp->w_doto = lused(lp);
#else
    curwp->w_dotp  = lp;
    curwp->w_doto  = 0;
#endif
    curwp->w_flag |= WFHARD;
    return (TRUE);
}

/*
 * This command is like "forwpage", but it goes backwards. overlap, like
 * above, is the overlap between the two windows. The value is from the ITS
 * EMACS manual. Bound to "M-V". We do a hard update for exactly the same
 * reason.
 */
int backpage(f, n)
     register int f;
     register int n;
{
    register LINE *lp;

#if BRIEF
	/*
 	 *	Brief allows the cursor after the end of a line.  It also
	 *	leaves the cursor on the same screen line on PgUp/PgDn
	 */
    int	line_offset = 0;

    /* find out how many lines into the screen is the dot */
    lp = curwp->w_linep;
    while(lp != curbp->b_linep && lp != curwp->w_dotp)
    {
        line_offset++;
	lp = lforw(lp);
    }
#endif

    if (f == FALSE)
    {
	n = curwp->w_ntrows - overlap;	/* Default scroll.      */
	if (n <= 0)		/* Don't blow up if the */
	    n = 1;		/* window is tiny.      */
    }
    else if (n < 0)
	return (forwpage(f, -n));
    lp = curwp->w_linep;
#if UNDOCMD
    if(!undo_in_progress)
        undo_save_dot(f, n);
#endif
    while (n-- && lback(lp) != curbp->b_linep)
	lp = lback(lp);
    curwp->w_linep = lp;

#if BRIEF
    while (line_offset-- && lp!=curbp->b_linep)
        lp = lforw(lp);
    curwp->w_dotp  = lp;
    if (curwp->w_doto > lused(lp))
        curwp->w_doto = lused(lp);
#else
    curwp->w_dotp  = lp;
    curwp->w_doto  = 0;
#endif

    curwp->w_flag |= WFHARD;
    return (TRUE);
}

/*
 * Set the mark in the current window to the value of "." in the window. No
 * errors are possible. Bound to "M-.".
 */
int setmark(f, n)
     int f,
         n;			/* argument flag and num */
{
/* make sure it is in range */
    if (f == FALSE)
	n = 0;
    n %= NMARKS;

#if UNDOCMD
    if(!undo_in_progress)
        undo_save_fn(FALSE, n, remmark);
#endif

    curwp->w_markp[n] = curwp->w_dotp;
    curwp->w_marko[n] = curwp->w_doto;
    mlwrite(TEXT9, n);
/*              "[Mark %d set]" */
    return (TRUE);
}

/*
 * Remove the mark in the current window.
 * Bound to ^X <space>
 */
int remmark(f, n)
     int f,
         n;			/* argument flag and num */
{
/* make sure it is in range */
    if (f == FALSE)
	n = 0;
    n %= NMARKS;

#if UNDOCMD
    if(!undo_in_progress)
        undo_save_fn(FALSE, n, setmark);
#endif

    curwp->w_markp[n] = NULL;
    curwp->w_marko[n] = 0;
    mlwrite(TEXT10, n);
/*              "[Mark %d removed]" */
    return (TRUE);
}

/*
 * Swap the values of "." and "mark" in the current window. This is pretty
 * easy, bacause all of the hard work gets done by the standard routine
 * that moves the mark about. The only possible error is "no mark". Bound to
 * "C-X C-X".
 */
int swapmark(f, n)
     int f,
         n;			/* argument flag and num */
{
    register LINE *odotp;
    register int odoto;

/* make sure it is in range */
    if (f == FALSE)
	n = 0;
    n %= NMARKS;

    if (curwp->w_markp[n] == NULL)
    {
	mlwrite(TEXT11, n);
    /*                      "No mark %d in this window" */
	return (FALSE);
    }
#if UNDOCMD
    if(!undo_in_progress)
        undo_save_dot(f, n);
#endif

    odotp = curwp->w_dotp;
    odoto = curwp->w_doto;
    curwp->w_dotp = curwp->w_markp[n];
    curwp->w_doto = curwp->w_marko[n];
    curwp->w_markp[n] = odotp;
    curwp->w_marko[n] = odoto;
    curwp->w_flag |= WFMOVE;
    return (TRUE);
}

/*
 * Goto a mark in the current window. This is pretty easy, bacause all of
 * the hard work gets done by the standard routine that moves the mark
 * about. The only possible error is "no mark". Bound to "M-^G".
 */
int gotomark(f, n)
     int f,
         n;			/* default and numeric args */
{
/* make sure it is in range */
    if (f == FALSE)
	n = 0;
    n %= NMARKS;

    if (curwp->w_markp[n] == NULL)
    {
	mlwrite(TEXT11, n);
    /*                      "No mark %d in this window" */
	return (FALSE);
    }
#if UNDOCMD
    if(!undo_in_progress)
        undo_save_dot(f, n);
#endif

    curwp->w_dotp = curwp->w_markp[n];
    curwp->w_doto = curwp->w_marko[n];
    curwp->w_flag |= WFMOVE;
    return (TRUE);
}

#if	DBCS
/* advance a char if we are on the second byte of a DBCS character */

int stopforw()
{
/* don't stop on the second byte of a 2 byte character */
    if (curwp->w_doto > 0 && is2byte(ltext(curwp->w_dotp),
				     ltext(curwp->w_dotp) + curwp->w_doto - 1))
	return (forwchar(TRUE, 1));
    return (TRUE);
}

/* retreat a char if we are on the second byte of a DBCS character */

int stopback()
{
/* don't stop on the second byte of a 2 byte character */
    if (curwp->w_doto > 0 && is2byte(ltext(curwp->w_dotp),
				     ltext(curwp->w_dotp) + curwp->w_doto - 1))
	return (backchar(TRUE, 1));
    return (TRUE);
}

#endif


#if GNMARKS
/* ============= Rtns supporting Multiple Named Marks ============== */

#define GNMARKCNT	10	/* no of marks in ring bfr */
#define MARKDEAD        -1	/* mark not accessable anymore */
typedef struct GNMARK
{
    short m_mname;		/* tag identifying mark */
    char m_bname[NBUFN];	/* Buffer mark assoc with */
    struct LINE *m_dotp;	/* ptr to line in buffer */
    short m_doto;		/* byte offset on line   */
}

GNMARK;
static GNMARK *nmarks[GNMARKCNT];	/* ptrs to mark structs - ring bfr */
static mmktop = 0;		/* next free mark ofset */

/*
 * get ptr to mark named 'mname', return NULL if not found
 */
static GNMARK **markget(mname)
     int mname;
{
    register GNMARK **mp;
    register int i;

    mp = nmarks;
    for (i = 0; *mp != NULL && i < GNMARKCNT; i++, mp++)
	if ((*mp)->m_mname == mname)	/* same name */
	    return mp;		/* return ptr to it */

    return NULL;		/* Failed */
}

/*
 * get name of mark fm par if set or keyboard otherwise
 */
static int getmarkname(f, n, prompt)
     int f,
         n;
     char *prompt;
{
    register int c;

    if (f == TRUE)
	return ((n % 10) + '0');


    if (clexec == FALSE)	/* interactive */
    {

	mlwrite(prompt);
	c = getkey();
	mlwrite("");		/* clear msg line */
	TTflush();
	if (c == abortc)
	{
	    ctrlg(FALSE, 0);
	    TTflush();
	    return (ABORT);
	}
    }
    else
    {
	char buf[20];
	int l;

	nextarg("", buf, 4, ctoec((int) '\r'));
    /*   quotes single char + 1 for eolchar */
	l = strlen(buf);
	if (buf[0] == '"' && buf[l - 1] == '"')
	    c = buf[1];		/* quoted char(s) */
	else
	    c = buf[0];
    }
    return c;
}

/* fix mark ptr if line referenced is being deleted  */
reset_nmark(lp, lpn)
     LINE *lp;
     LINE *lpn;
{
    register GNMARK **mp;
    register int i;

    mp = nmarks;
    for (i = 0; *mp != NULL && i < GNMARKCNT; i++, mp++)
	if (lp == (*mp)->m_dotp)/* same line */
	{
	    (*mp)->m_dotp = lpn;/* set to next line */
	    (*mp)->m_doto = 0;	/* ofs 0 */
	}
}

/*
 * make a named mark from 'n' if set ( mapped onto name '0'- '9' )
 * otherwise prompt for name.
 * return TRUE if made mark OK, FALSE if failed, ABORT if chged mind
 * Bound to M-,
 */
setnmark(f, n)
{
    register int c;
    register GNMARK **mp;

    if ((c = getmarkname(f, n, "Make Mark:")) == ABORT)
	return ABORT;

    if ((mp = markget(c)) == NULL)	/* markname not already exist */
    {
	if ((mp = markget(MARKDEAD)) == NULL)	/* no freed marks */
	{			/* so Overwrite top of mark stack */
	    mp = nmarks + mmktop;
	    if (*mp == NULL)	/* mark structure not allocated ?? */
	    {			/* allocate it */
		if ((*mp = (GNMARK *) malloc(sizeof(GNMARK))) == NULL)
		{
		    mlwrite("Cannot malloc space for New Mark");
		    return FALSE;
		}
	    }
	}
	mmktop++;
	mmktop %= GNMARKCNT;
    }

    (*mp)->m_mname = c;		/* mark name */
    strcpy((*mp)->m_bname, curbp->b_bname);	/* buffer name */
    (*mp)->m_dotp = curwp->w_dotp;	/* 'dot' line */
    (*mp)->m_doto = curwp->w_doto;	/* 'dot' ofset */

    return TRUE;
}

/*
 * Jump to mark named 'n' if set otherwise prompt for name
 * Returns TRUE if got to mark O.K, FALSE if no mark of that name or
 * buffer no longer accessable
 */
jmpnmark(f, n)
{
    register int c;
    register GNMARK **mp;
    register BUFFER *bp;


    if ((c = getmarkname(f, n, "Jmp to Mark:")) == ABORT)
	return ABORT;

    if ((mp = markget(c)) == NULL)
    {
	mlwrite("Mark not set");
	return FALSE;
    }
/*
 *   mlwrite("Mark %c: Bfr %s ofst %d",
 *        (int)(*mp)->m_mname, (*mp)->m_bname, (int)(*mp)->m_doto);
 */

    if ((bp = bfind((*mp)->m_bname, FALSE, 0)) == NULL)
    {
	mlwrite("Buffer out of scope");
	(*mp)->m_mname = MARKDEAD;
	return FALSE;
    }

    if (bp != curbp)
	swbuffer(bp);		/* chg buffers */
    else
	curwp->w_flag |= WFFORCE | WFHARD;	/* Force scn update */

    curwp->w_dotp = (*mp)->m_dotp;
    curwp->w_doto = (*mp)->m_doto;
    return TRUE;
}

/* ============= End of Rtns supporting Multiple Named Marks ============== */
#endif /* GNMARKS */
