/*	EVAR.H: Environment and user variable definitions
 *		for MicroEMACS
 *
 *		written 1993 by Daniel Lawrence
 *
 *        Converted internal structures for env var and  functions to be
 *        name/token value pair to make it simpler to add vars/fns.
 *        Compensating mods also in eval.c    27.06.93 Hops.
 */

/* ---- Structure to hold user variables and their definitions ---- */

typedef struct UVAR
{
    char u_name[NVSIZE + 1];	/* name of user variable */
    char *u_value;		/* value (string) */
}
UVAR;

/*	current user variables (This structure will probably change)	*/

#define MAXVARS 	512

UVAR uv[MAXVARS + 1];	/* user variables */


/* ---------------- List of recognized environment variables ---------------- */

/* token names 	*/

#define EVACOUNT	0
#define EVASAVE         1
#define	EVBUFHOOK	2
#define EVCBFLAGS	3
#define EVCBUFNAME	4
#define EVCFNAME	5
#define EVCMDHK 	6
#define EVCMODE 	7
#define EVCURCHAR	8
#define EVCURCOL	9
#define EVCURLINE	10
#define EVCURWIDTH	11
#define	EVCURWIND	12
#define EVCWLINE	13
#define EVDEBUG         14
#define	EVDESKCLR	15
#define EVDIAGFLAG      16
#define EVDISCMD        17
#define EVDISINP        18
#define	EVDISPHIGH	19
#define EVEXBHOOK       20
#define EVFCOL		21
#define EVFILLCOL	22
#define EVFLICKER	23
#define	EVFMTLEAD	24
#define EVGFLAGS	25
#define EVGMODE 	26
#define	EVHARDTAB	27
#define	EVHILITE	28
#define EVHJUMP		29
#define EVHSCRLBAR      30
#define EVHSCROLL	31
#define EVISTERM	32
#define EVKILL          33
#define EVLANG          34
#define EVLASTKEY       35
#define EVLASTMESG      36
#define EVLINE          37
#define	EVLTERM		38
#define EVLWIDTH        39
#define EVMATCH         40
#define EVMMOVE		41
#define EVMODEFLAG      42
#define EVMSFLAG        43
#define	EVNUMWIND	44
#define	EVORGCOL	45
#define	EVORGROW	46
#define	EVOS		47
#define	EVOVERLAP	48
#define EVPAGELEN       49
#define EVPALETTE       50
#define	EVPARALEAD	51
#define EVPENDING       52
#define	EVPOPFLAG	53
#define	EVPOSFLAG	54
#define EVPROGNAME      55
#define EVRAM           56
#define EVREADHK        57
#define	EVREGION	58
#define EVREPLACE       59
#define EVRVAL          60
#define EVSCRNAME	61
#define EVSEARCH        62
#define EVSEARCHPNT	63
#define EVSEED          64
#define EVSOFTTAB	65
#define EVSRES          66
#define EVSSAVE         67
#define EVSSCROLL	68
#define EVSTATUS	69
#define EVSTERM 	70
#define EVTARGET	71
#define EVTIME		72
#define EVTIMEFLAG	73
#define EVTPAUSE	74
#define EVVERSION	75
#define EVVSCRLBAR      76
#define	EVWCHARS	77
#define EVWLINE 	78
#define EVWRAPHK	79
#define	EVWRITEHK	80
#define EVXPOS		81
#define	EVYANKFLAG	82
#define EVYPOS		83

#if NEWVAR
#define EVDATE          84
#define EVCHAR          85
#define EVCUROFS        86
#define EVUARG          87
#define EVMARK          88
#define EVMCOL          89
#define EVMLINE         90
#define EVUCONFIG       91
#define EVEXCLSFX       92
#endif

/* Env name/token list - Keep in alpha order, Names should be <= 8 chars */

typedef struct EVAR
{
    char *e_name;		/* name of env variable */
    unsigned char e_token;	/* token to return for it */
}
EVAR;

EVAR envars[] =
{
    {"acount",  EVACOUNT},	/* # of chars until next auto-save */
    {"asave",   EVASAVE},	/* # of chars between auto-saves */
    {"bufhook", EVBUFHOOK},	/* enter buffer switch hook */
    {"cbflags", EVCBFLAGS},	/* current buffer flags */
    {"cbufname", EVCBUFNAME},	/* current buffer name */
    {"cfname",  EVCFNAME},	/* current file name */
#if NEWVAR
    {"char",    EVCHAR},	/* string val of char at point */
#endif /* NEWVAR */
    {"cmdhook", EVCMDHK},	/* command loop hook */
    {"cmode",   EVCMODE},	/* mode of current buffer */
    {"curchar", EVCURCHAR},	/* current character under the cursor */
    {"curcol",  EVCURCOL},	/* current column pos of cursor */
    {"curline", EVCURLINE},	/* current line in file */
#if NEWVAR
    {"curofs",   EVCUROFS},	/* ofset of point on line */
#endif /* NEWVAR */
    {"curwidth", EVCURWIDTH},	/* current screen width */
    {"curwind",  EVCURWIND},	/* current window ordinal on current screen */
    {"cwline",   EVCWLINE},	/* current screen line in window */
#ifdef NEWVAR
    {"date",     EVDATE},	/* Date/time in YYMMDDHHMM format  */
#endif
    {"debug",    EVDEBUG},	/* macro debugging */
    {"deskcolor",EVDESKCLR},	/* desktop color */
    {"diagflag", EVDIAGFLAG},	/* diagonal mouse movements enabled? */
    {"discmd",   EVDISCMD},	/* display commands on command line */
    {"disinp",   EVDISINP},	/* display command line input characters */
    {"disphigh", EVDISPHIGH},	/* display high bit characters escaped */
    {"exbhook",  EVEXBHOOK},	/* exit buffer switch hook */
#ifdef NEWVAR
    {"exclsfx",  EVEXCLSFX},	/* filename exclusion suffix list */
#endif
    {"fcol",     EVFCOL},	/* first displayed column in curent window */
    {"fillcol",  EVFILLCOL},	/* current fill column */
    {"flicker",  EVFLICKER},	/* flicker supression */
    {"fmtlead",  EVFMTLEAD},	/* format command lead characters */
    {"gflags",   EVGFLAGS},	/* global internal emacs flags */
    {"gmode",    EVGMODE},	/* global modes */
    {"hardtab",  EVHARDTAB},	/* current hard tab size */
    {"hilight",  EVHILITE},	/* region # to hilight (255 to turn off) */
    {"hjump",    EVHJUMP},	/* horizontal screen jump size */
    {"hscrlbar", EVHSCRLBAR},	/* horizontal scroll bar flag */
    {"hscroll",  EVHSCROLL},	/* horizontal scrolling flag */
    {"isterm",   EVISTERM},	/* incremental-search terminator character */
    {"kill",     EVKILL},	/* kill buffer (read only) */
    {"language", EVLANG},	/* language of text messages */
    {"lastkey",  EVLASTKEY},	/* last keyboard char struck */
    {"lastmesg", EVLASTMESG},	/* last string mlwrite()ed */
    {"line",     EVLINE},	/* text of current line */
    {"lterm",    EVLTERM},	/* current line terminator for writes */
    {"lwidth",   EVLWIDTH},	/* width of current line */
#if NEWVAR
    {"mark",     EVMARK},	/* TRUE if mark set in current window */
    {"markline", EVMLINE},	/* current line of mark in current buffer */
    {"markofs",  EVMCOL},	/* column ofs of mark in current buffer */
#endif /* NEWVAR */
    {"match",    EVMATCH},	/* last matched magic pattern */
    {"mmove",    EVMMOVE},	/* mouse moves events style */
    {"modeflag", EVMODEFLAG},	/* Modelines displayed flag */
    {"msflag",   EVMSFLAG},	/* activate mouse? */
    {"numwind",  EVNUMWIND},	/* number of windows on current screen */
    {"orgcol",   EVORGCOL},	/* screen origin column */
    {"orgrow",   EVORGROW},	/* screen origin row */
    {"os",       EVOS},		/* what Operating System? */
    {"overlap",  EVOVERLAP},	/* line overlap when paging */
    {"pagelen",  EVPAGELEN},	/* number of lines used by editor */
    {"palette",  EVPALETTE},	/* current palette string */
    {"paralead", EVPARALEAD},	/* paragraph leadin characters */
    {"pending",  EVPENDING},	/* type ahead pending flag */
    {"popflag",  EVPOPFLAG},	/* pop-up windows active? */
    {"posflag",  EVPOSFLAG},	/* display point position on modeline? */
    {"progname", EVPROGNAME},	/*  current prog name - "MicroEMACS" */
    {"ram",      EVRAM},	/* ram in use by malloc */
    {"readhook", EVREADHK},	/* read file execution hook */
    {"region",   EVREGION},	/* current region (read only) */
    {"replace",  EVREPLACE},	/* replacement pattern */
    {"rval",     EVRVAL},	/* child process return value */
    {"scrname",  EVSCRNAME},	/* current screen name */
    {"search",   EVSEARCH},	/* search pattern */
    {"searchpnt",EVSEARCHPNT},	/* differing search styles (term point) */
    {"seed",     EVSEED},	/* current random number seed */
    {"softtab",  EVSOFTTAB},	/* current soft tab size */
    {"sres",     EVSRES},	/* current screen resolution */
    {"ssave",    EVSSAVE},		/* safe save flag */
    {"sscroll",  EVSSCROLL},	/* smooth scrolling flag */
    {"status",   EVSTATUS},	/* returns the status of the last command */
    {"sterm",    EVSTERM},	/* search terminator character */
    {"target",   EVTARGET},	/* target for line moves */
    {"time",     EVTIME},	/* date and time */
    {"timeflag", EVTIMEFLAG},	/* display time? */
    {"tpause",   EVTPAUSE},	/* length to pause for paren matching */
#ifdef NEWVAR
    {"uarg",      EVUARG},	/* current value of universal arg  */
    {"uconfig",   EVUCONFIG},	/* flag for some usr config items  */
#endif /* NEWVAR */
    {"version",   EVVERSION},	/* current version number */
    {"vscrlbar",  EVVSCRLBAR},	/* vertical scroll bar flag */
    {"wchars",    EVWCHARS},	/* set of characters legal in words */
    {"wline",     EVWLINE},	/* # of lines in current window */
    {"wraphook",  EVWRAPHK},	/* wrap word execution hook */
    {"writehook", EVWRITEHK},	/* write file hook */
    {"xpos",      EVXPOS},	/* current mouse X position */
    {"yankflag",  EVYANKFLAG},	/* point placement at yanked/incl text */
    {"ypos",      EVYPOS}	/* current mouse Y position */
};

#define NEVARS	sizeof(envars) / sizeof(char *)

/* ---------------- List of recognized user functions ---------------- */

typedef struct UFUNC
{
    char *f_name;		/* name of function */
    int f_type;			/* 1 = monamic, 2 = dynamic */
    unsigned char f_token;	/* fn token value */
}
UFUNC;

/* Arg type defns */
#define NILNAMIC	0
#define MONAMIC 	1
#define DYNAMIC 	2
#define TRINAMIC	3

/*	Token definitions 	*/

#define UFABS		0
#define UFADD		1
#define UFAND		2
#define UFASCII 	3
#define UFBAND		4
#define UFBIND		5
#define UFBNOT		6
#define UFBOR		7
#define UFBXOR		8
#define UFCAT		9
#define UFCHR		10
#define UFDIV		11
#define UFENV		12
#define UFEQUAL 	13
#define UFEXIST 	14
#define UFFIND		15
#define UFGREATER	16
#define UFGROUP		17
#define UFGTCMD 	18
#define UFGTKEY 	19
#define UFIND		20
#define	UFISNUM		21
#define UFLEFT		22
#define UFLENGTH	23
#define UFLESS		24
#define UFLOWER 	25
#define UFMID		26
#define UFMOD		27
#define UFNEG		28
#define UFNOT		29
#define UFOR		30
#define UFREVERSE	31
#define UFRIGHT 	32
#define UFRND		33
#define UFSEQUAL	34
#define UFSGREAT	35
#define UFSINDEX	36
#define UFSLESS 	37
#define	UFSLOWER	38
#define UFSUB		39
#define	UFSUPPER	40
#define UFTIMES 	41
#define	UFTRIM		42
#define UFTRUTH 	43
#define UFUPPER 	44
#define UFXLATE 	45
#define UFSRHS          46
#if NEWVAR
#define UFSBREAK        47
#endif /* NEWVAR */

UFUNC funcs[] =
{
    "abs", MONAMIC, UFABS,	/* absolute value of a number */
    "add", DYNAMIC, UFADD,	/* add two numbers together */
    "and", DYNAMIC, UFAND,	/* logical and */
    "asc", MONAMIC, UFASCII,	/* char to integer conversion */
    "ban", DYNAMIC, UFBAND,	/* bitwise and	 9-10-87  jwm */
    "bin", MONAMIC, UFBIND,	/* loopup what function name is bound to a key */
    "bno", MONAMIC, UFBNOT,	/* bitwise not */
    "bor", DYNAMIC, UFBOR,	/* bitwise or	 9-10-87  jwm */
    "bxo", DYNAMIC, UFBXOR,	/* bitwise xor	 9-10-87  jwm */
    "cat", DYNAMIC, UFCAT,	/* concatinate string */
    "chr", MONAMIC, UFCHR,	/* integer to char conversion */
    "div", DYNAMIC, UFDIV,	/* division */
    "env", MONAMIC, UFENV,	/* retrieve a system environment var */
    "equ", DYNAMIC, UFEQUAL,	/* logical equality check */
    "exi", MONAMIC, UFEXIST,	/* check if a file exists */
    "fin", MONAMIC, UFFIND,	/* look for a file on the path... */
    "gre", DYNAMIC, UFGREATER,	/* logical greater than */
    "gro", MONAMIC, UFGROUP,	/* return group match in MAGIC mode */
    "gtc", NILNAMIC, UFGTCMD,	/* get 1 emacs command */
    "gtk", NILNAMIC, UFGTKEY,	/* get 1 charater */
    "ind", MONAMIC, UFIND,	/* evaluate indirect value */
    "isn", MONAMIC, UFISNUM,	/* is the arg a number? */
    "lef", DYNAMIC, UFLEFT,	/* left string(string, len) */
    "len", MONAMIC, UFLENGTH,	/* string length */
    "les", DYNAMIC, UFLESS,	/* logical less than */
    "low", MONAMIC, UFLOWER,	/* lower case string */
    "mid", TRINAMIC, UFMID,	/* mid string(string, pos, len) */
    "mod", DYNAMIC, UFMOD,	/* mod */
    "neg", MONAMIC, UFNEG,	/* negate */
    "not", MONAMIC, UFNOT,	/* logical not */
    "or",  DYNAMIC, UFOR,	/* logical or */
    "rev", MONAMIC, UFREVERSE,	/* reverse */
    "rhs", DYNAMIC, UFSRHS,	/* right string(string, n) - rhs n chars - NU */
/* Above should rplc right but not backward  */
/* compatible so give new name .... Hops */
    "rig", DYNAMIC, UFRIGHT,	/* right string(s, pos) - chars fm pos up OLD */
    "rnd", MONAMIC, UFRND,	/* get a random number */
#if NEWVAR
    "sbr", DYNAMIC, UFSBREAK,	/* index of any char in string 2 in string 1 */
#endif /* NEWVAR */
    "seq", DYNAMIC, UFSEQUAL,	/* string logical equality check */
    "sgr", DYNAMIC, UFSGREAT,	/* string logical greater than */
    "sin", DYNAMIC, UFSINDEX,	/* find the index of one string in another */
    "sle", DYNAMIC, UFSLESS,	/* string logical less than */
    "slo", DYNAMIC, UFSLOWER,	/* set lower to upper char translation */
    "sub", DYNAMIC, UFSUB,	/* subtraction */
    "sup", DYNAMIC, UFSUPPER,	/* set upper to lower char translation */
    "tim", DYNAMIC, UFTIMES,	/* multiplication */
    "tri", MONAMIC, UFTRIM,	/* trim whitespace off the end of a string */
    "tru", MONAMIC, UFTRUTH,	/* Truth of the universe logical test */
    "upp", MONAMIC, UFUPPER,	/* uppercase string */
    "xla", TRINAMIC, UFXLATE	/* XLATE character string translation */
};

#define NFUNCS	sizeof(funcs) / sizeof(UFUNC)
