/*	Input:	Various input routines for MicroEMACS
		written by Daniel Lawrence
		copyright 1986-1993 by Daniel M. Lawrence

	Notes:

	MicroEMACS's kernel processes two distinct forms of
	characters.  One of these is a standard unsigned character
	which is used in the edited text.  The other form, called
	an EMACS Extended Character is a 2 byte value which contains
	both an ascii value, and flags for certain prefixes/events.

	Bit	Usage
	---	-----
	0 -> 7	Standard 8 bit ascii character
	8	Control key flag
	9	META prefix flag
	10	^X prefix flag
	11	Function key flag
	12	Mouse prefix
	13	Shifted flag (not needed on alpha shifted characters)
	14	Alterate prefix (ALT key on PCs)

	The machine dependent driver is responsible for returning
	a byte stream from the various input devices with various
	prefixes/events embedded as escape codes.  Zero is used as the
	value indicating an escape sequence is next.  The format of
	an escape sequence is as follows:

	0		Escape indicator
	<prefix byte>	upper byte of extended character
	{<col><row>}	col, row position if the prefix byte
			indicated a mouse event or a menu selection
			in which case these form a 16 bit menu ID
	<event code>	value of event

	A ^<space> sequence (0/1/32) is generated when an actual
	null is being input from the control-space key under many
	unix systems.  These values are then interpreted by getkey()
	to construct the proper extended character sequences to pass
	to the MicroEMACS kernel.
*/

#include	<stdio.h>
#include	"estruct.h"
#include	"eproto.h"
#include	"edef.h"
#include	"elang.h"

#if ENVFUNC && (USG | AUX | BSD | V7 | SUN | HPUX8 | HPUX9)
#include        <stdlib.h>   /* for getenv() */
#include	<pwd.h>
extern struct passwd *getpwnam();

#endif

/* define local macros for enabling new goodies */
#define MSG_STACK   1
#define INTERNAL_METACHARS    1
#define TMP_COMPCHAR 0xff    /* pseudo char value to trigger where completion
                             * should be done, completion keys and cases get
                             * mapped to this value
                             */

#ifndef cntl
#define cntl(x)		(x & 0x1f)
#endif

/*
 * Ask a yes or no question in the message line. Return either TRUE, FALSE, or
 * ABORT. The ABORT status is returned if the user bumps out of the question
 * with a ^G. Used any time a confirmation is required.
 * Modified to add return code of FAILED. Used to indicate to when exiting
 * with dirty buffers to write them all out before exiting. Hops 19.06.93
 */

#if	!WINDOW_MSWIN	     /* for MS Windows, mlyesno defined in mswsys.c */
int mlyesno(prompt)
     char *prompt;
{
    int c;		     /* input character */
    char buf[NPAT];	     /* prompt to user */

#if EXIT_WRITE
    int ew_flag = (strncmp(prompt, TEXT104, strlen(TEXT104)) == 0);

    /* modifying correct message ???*/
#endif

    for (;;)
    {
	/* build and prompt the user */
	strcpy(buf, prompt);
#if EXIT_WRITE
	strcat(buf, (!ew_flag) ? TEXT162 : " [y/n/w] ");
#else
	strcat(buf, TEXT162);	/*  " [y/n]? " */
#endif
	mlwrite(buf);

	/* get the response */
	c = getcmd();
	/* getcmd() lets us check for anything that might generate
	         * a 'y' or 'Y' in case use screws up
	         */

	if (c == ectoc(abortc))	/* Bail out! */
	    return (ABORT);

	if ((c == 'n') || (c == 'N')
	    || (c & (SPEC | ALTD | CTRL | META | CTLX | MOUS)))
	    return (FALSE);	/* ONLY 'y' or 'Y' allowed!!! */

#if	FRENCH
	if (c == 'o' || c == 'O')
	    return (TRUE);
#endif

	if (c == 'y' || c == 'Y')
	    return (TRUE);

#if EXIT_WRITE
	/* special for modified buffers prompt */
	if (ew_flag && c == 'w' || c == 'W')
	    return (FAILED);	/* means write everything out */
#endif
	return (FALSE);
    }
}

#endif

/*
 * Write a prompt into the message line, then read back a response. Keep
 * track of the physical position of the cursor. If we are in a keyboard
 * macro throw the prompt away, and return the remembered response. This
 * lets macros run at full speed. The reply is always terminated by a carriage
 * return. Handle erase, kill, and abort keys.
 */
int mlreply(prompt, buf, nbuf)
     char *prompt;
     char *buf;
     int nbuf;
{
    return (nextarg(prompt, buf, nbuf, ctoec((int) '\r')));
}

/*
 *	ectoc:	expanded character to character
 *	collapse the CTRL and SPEC flags back into an ascii code   
 */

int ectoc(c)
     int c;
{
    if (c == (CTRL | ' '))
	c = 0;
    if (c & CTRL)
	c = c ^ (CTRL | 0x40);
    if (c & SPEC)
	c = c & 255;
    return (c);
}

/*
 *	ctoec:	character to extended character
 *		pull out the CTRL and SPEC prefixes (if possible)	
 */

int ctoec(c)
     int c;
{
    if (c == 0)
	c = CTRL | ' ';
    else if ((c >= 0x00 && c <= 0x1F) || c == 0x7F)
	c = CTRL | (c ^ 0x40);
    return (c);
}

/* 
 * Get a command name from the command line. Command completion means
 *  that pressing a <SPACE> will attempt to complete an unfinished command
 *  name if it is unique.
 */

#if	MSC
int (*  getname(char *prompt)) (void)
#else
int (*  getname(prompt)) ()
     char *prompt;	     /* string to prompt with */
#endif
{
    char *sp;		     /* ptr to the returned string */

    sp = complete(prompt, NULL, CMP_COMMAND, NSTRING);
    if (sp == NULL)
	return (NULL);

    return (fncmatch(sp));
}

/*	getcbuf:	get a completion from the user for a buffer name.
 *
 *	I was goaded into this by lots of other people's
 *	completion code.
 */

BUFFER *getcbuf(prompt, defval, createflag)
     char *prompt;	     /* prompt to user on command line */
     char *defval;	     /* default value to display to user */
     int createflag;	     /* should this create a new buffer? */
{
    char *sp;		     /* ptr to the returned string */

    sp = complete(prompt, defval, CMP_BUFFER, NBUFN);
    if (sp == NULL)
	return (NULL);

    return (bfind(sp, createflag, 0));
}

char *gtfilename(prompt)
     char *prompt;	     /* prompt to user on command line */
{
#if	MSDOS | OS2
    char *scan;

#endif
#if	WINDOW_MSWIN
    static char sp[NFILEN];

    if (!FILENAMEREPLY(prompt, sp, NFILEN))
	return NULL;
#else
    char *sp;		     /* ptr to the returned string */

    /* get a file name, default to current buffer's for 'Read File' only */
    if (strcmp(prompt, TEXT131) == 0 &&
	curbp && curbp->b_fname && strcmp(curbp->b_fname, "") != 0)
	sp = complete(prompt, curbp->b_fname, CMP_FILENAME, NFILEN);
    else
	sp = complete(prompt, NULL, CMP_FILENAME, NFILEN);
#endif
#if	MSDOS | OS2
    /* change forward slashes to back */
    if (sp)
    {
	scan = sp;
	while (*scan)
	{
	    if (*scan == '/')
		*scan = DIRSEPCHAR;
	    ++scan;
	}
    }
#endif
    return (sp);
}

/* make up a prompt string suitable for passing onto getstring()  */
char *mkprompt(prompt, defval, iterm)
     char *prompt;
     char *defval;
     int iterm;
{
    static char promptbuf[NSTRING];
    char buf[NSTRING];	     /* buffer to hold tentative name */

    strcpy(promptbuf, prompt);
    if (defval != NULL && *defval != '\0')
    {
	strcat(promptbuf, " [");
	strcat(promptbuf, defval);
	strcat(promptbuf, "]");	/* "%s[%s]", prompt, defval */
    }

    /* Display the proper current search terminator character. */
    if (iterm == (CTRL | '@') /* || iterm == (CTRL | 'M')  */ )
    {
	strcat(promptbuf, ": ");
    }
    else
    {
	strcat(promptbuf, "<");
	switch (iterm)
	{
	case CTRL | '[':
	    strcat(promptbuf, "META");
	    break;
	case CTRL | 'M':
	    strcat(promptbuf, "NL");
	    break;
	default:
	    strcat(promptbuf, cmdstr(iterm, buf));
	    break;
	}
	strcat(promptbuf, ">: ");
    }

    return promptbuf;
}


/*
 * Entry Routine for the various types of completion
 */
char *complete(prompt, defval, type, maxlen)
     char *prompt;	     /* prompt to user on command line */
     char *defval;	     /* default value to display to user */
     int type;		     /* type of what we are completing */
     int maxlen;	     /* maximum length of input field */
{
    register int c;	     /* current input character */
    register int ec;	     /* extended input character */
    static char buf[NSTRING];/* buffer to hold tentative name */
    int r;
    char aprompt[NSTRING],
        *promptp;


    /* if we are executing a command line get the next arg and match it */
    if (clexec)
    {
	if (macarg(buf) != TRUE)
	    return (NULL);
	return (buf);
    }

    /* if it exists, prompt the user for a buffer name */
    promptp = prompt;
    if (prompt)
    {
	if (type == CMP_COMMAND)
	    ;
	else
	    promptp = mkprompt(prompt, defval, sterm);
    }
    r = getstring(promptp, buf, maxlen, ctoec(RETCHAR), type);
    if (r == FALSE && defval)	/* handle default */
    {
	strcpy(buf, defval);
	r = TRUE;
    }
    return r == TRUE ? buf : NULL;
}



/*	comp_command:	Attempt a completion on a command name	*/

int comp_command(name, cpos)
     char *name;	     /* comd containing the current name to complete */
     int *cpos;		     /* ptr to position of next character to insert */
{
    register NBIND *bp;	     /* trial command to complete */
    register int index;	     /* index into strings to compare */
    register int curbind;    /* index into the names[] array */
    register NBIND *match;   /* last command that matches string */
    register int matchflag;  /* did this command name match? */
    register int comflag;    /* was there a completion at all? */

    /* start attempting completions, one character at a time */
    comflag = FALSE;
    curbind = 0;
    while (*cpos < NSTRING)
    {

	/* first, we start at the first command and scan the list */
	match = NULL;
	curbind = 0;
	while (curbind <= numfunc)
	{

	    /* is this a match? */
	    bp = &names[curbind];
	    matchflag = TRUE;
	    for (index = 0; index < *cpos; index++)
		if (name[index] != bp->n_name[index])
		{
		    matchflag = FALSE;
		    break;
		}

	    /* if it is a match */
	    if (matchflag)
	    {

		/* if this is the first match, simply record it */
		if (match == NULL)
		{
		    match = bp;
		    name[*cpos] = bp->n_name[*cpos];
		}
		else
		{
		    /* if there's a difference, stop here */
		    if (name[*cpos] != bp->n_name[*cpos])
		    {
			clist_command(name, cpos);
			update(TRUE);
			return FALSE;
		    }
		}
	    }

	    /* on to the next command */
	    curbind++;
	}

	/* with no match, we are done */
	if (match == NULL)
	{
	    /* beep if we never matched */
	    if (comflag == FALSE)
		TTbeep();
	    return comflag;
	}

	/* if we have completed all the way... go back */
	if (name[*cpos] == 0)
	{
	    return comflag;
	}

	/* remember we matched, and complete one character */
	comflag = TRUE;
	TTputc(name[(*cpos)++]);
	TTflush();
    }

    /* don't allow a completion past the end of the max command name length */
    return comflag;
}

/*	clist_command:	Make a completion list based on a partial name */
int clist_command(name, cpos)
     char *name;	     /* command containing the current name to complete */
     int *cpos;		     /* ptr to position of next character to insert */
{
    register NBIND *bp;	     /* trial command to complete */
    register int curbind;    /* index into the names[] array */
    register int name_len;   /* current length of input string */
    register BUFFER *listbuf;/* buffer to put completion list into */

    /* get a buffer for the completion list */
    listbuf = bfind("[Completion list]", TRUE, BFINVS);
    if (listbuf == NULL || bclear(listbuf) == FALSE)
    {
	ctrlg(FALSE, 0);
	TTflush();
	return;
    }

    name_len = *cpos;

    /* first, we start at the first command and scan the list */
    for (curbind = 0; curbind <= numfunc; curbind++)
    {

	/* is this a match? */
	bp = &names[curbind];
	if (strncmp(name, bp->n_name, name_len) == 0)
	    addline(listbuf, bp->n_name);
    }

    wpopup(listbuf);
    return;
}

/*	comp_buffer:	Attempt a completion on a buffer name	*/
comp_buffer(name, cpos)
     char *name;	     /* buffer containing the current name to complete */
     int *cpos;		     /* ptr to position of next character to insert */
{
    register BUFFER *bp;     /* trial buffer to complete */
    register int index;	     /* index into strings to compare */
    register BUFFER *match;  /* last buffer that matches string */
    register int matchflag;  /* did this buffer name match? */
    register int comflag;    /* was there a completion at all? */

    /* start attempting completions, one character at a time */
    comflag = FALSE;
    while (*cpos < NBUFN)
    {

	/* first, we start at the first buffer and scan the list */
	match = NULL;
	bp = bheadp;
	while (bp)
	{

	    /* is this a match? */
	    matchflag = TRUE;
	    for (index = 0; index < *cpos; index++)
		if (name[index] != bp->b_bname[index])
		{
		    matchflag = FALSE;
		    break;
		}

	    /* if it is a match */
	    if (matchflag)
	    {

		/* if this is the first match, simply record it */
		if (match == NULL)
		{
		    match = bp;
		    name[*cpos] = bp->b_bname[*cpos];
		}
		else
		{
		    /* if there's a difference, stop here */
		    if (name[*cpos] != bp->b_bname[*cpos])
			return;
		}
	    }

	    /* on to the next buffer */
	    bp = bp->b_bufp;
	}

	/* with no match, we are done */
	if (match == NULL)
	{
	    /* beep if we never matched */
	    if (comflag == FALSE)
		TTbeep();
	    return;
	}

	/* if we have completed all the way... go back */
	if (name[*cpos] == 0)
	{
	    return;
	}

	/* remember we matched, and complete one character */
	comflag = TRUE;
	TTputc(name[(*cpos)++]);
	TTflush();
    }

    /* don't allow a completion past the end of the max buffer name length */
    return;
}

/* clist_buffer:    Make a completion list based on a partial buffer name */
static void clist_buffer(name, cpos)
     char *name;	     /* comd containing the current name to complete */
     int *cpos;		     /* ptr to position of next character to insert */
{
    register int name_len;   /* current length of input string */
    register BUFFER *listbuf;/* buffer to put completion list into */
    register BUFFER *bp;     /* trial buffer to complete */

    /* get a buffer for the completion list */
    listbuf = bfind("[Completion list]", TRUE, BFINVS);
    if (listbuf == NULL || bclear(listbuf) == FALSE)
    {
	ctrlg(FALSE, 0);
	TTflush();
	return;
    }

    /* first, we start at the first buffer and scan the list */
    name_len = *cpos;
    bp = bheadp;

    while (bp)
    {

	/* is this a match? */
	if (strncmp(name, bp->b_bname, name_len) == 0)
	    addline(listbuf, bp->b_bname);

	/* on to the next buffer */
	bp = bp->b_bufp;
    }

    wpopup(listbuf);
    return;
}

#if	!WINDOW_MSWIN

/* Test if name is in file exclusion list.
 * Current list is hard coded ( should be in macro var ) and tests only on
 * suffixes
 */

static int file_excluded( fname )
    char *fname ;
{
    char *s;
    
    if (!excl_sfxs || ! *excl_sfxs )        /* no excluded suffixes */
        return FALSE;
        
    if ( (s = strrchr(fname, '.')) == NULL )    /* no extension */
        return FALSE;

    return  ( strstr(excl_sfxs, s ) != NULL ) ;
}
    
/*	clist_file:	Make a completion list based on a partial file name */
static void clist_file(name, cpos)
     char *name;	     /* comd containing the current name to complete */
     int *cpos;		     /* ptr to position of next character to insert */

{
    register int name_len;   /* current length of input string */
    register BUFFER *listbuf;/* buffer to put completion list into */
    register char *fname;    /* trial file to complete */

    /* get a buffer for the completion list */
    listbuf = bfind("[Completion list]", TRUE, BFINVS);
    if (listbuf == NULL || bclear(listbuf) == FALSE)
    {
	ctrlg(FALSE, 0);
	TTflush();
	return;
    }

    /* first, we start at the first file and scan the list */
    name_len = *cpos;
    name[*cpos] = 0;
    fname = getffile(name);

    /* first, we start at the first file and scan the list */
    while (fname)
    {

	/* is this a match? */
	if (strncmp(name, fname, name_len) == 0 && ! file_excluded(fname) )
	    addline(listbuf, fname);

	/* on to the next file */
	fname = getnfile();
    }

    wpopup(listbuf);
    return;
}

/*
 *	comp_file:	Attempt a completion on a file name
 *      return TRUE if match to a single item FALSE otherwise
 */

int comp_file(name, cpos)
     char *name;	     /* file containing the current name to complete */
     int *cpos;		     /* ptr to position of next character to insert */
{
    register char *fname;    /* trial file to complete */
    register int index;	     /* index into strings to compare */

    register int matches;    /* number of matches for name */
    char longestmatch[NSTRING];	/* temp buffer for longest match */
    int longestlen;	     /* length of longest match (always > *cpos) */
    char *matchname;
    int retval = FALSE;

    /* everything (or nothing) matches an empty string */
    if (*cpos == 0)
	return retval;

    /* first, we start at the first file and scan the list */
    matches = 0;
    name[*cpos] = 0;
    fname = getffile(name);
    while (fname)
    {

	/* is this a match? */
	if (strncmp(name, fname, *cpos) == 0 && !file_excluded( fname) )
	{

	    /* count the number of matches */
	    matches++;

	    /* if this is the first match, simply record it */
	    if (matches == 1)
	    {
		strcpy(longestmatch, fname);
		longestlen = strlen(longestmatch);
	    }
	    else
	    {

		/* if there's a difference, stop here */
		/* -> multiple possible matches */
		if (longestmatch[*cpos] != fname[*cpos])
		    goto do_clist;

		for (index = (*cpos) + 1; index < longestlen; index++)
		    if (longestmatch[index] != fname[index])
		    {
			longestlen = index;
			longestmatch[longestlen] = 0;
		    }
	    }
	}

	/* on to the next file */
	fname = getnfile();
    }

    /* beep if we never matched */
    if (matches == 0)
    {
	TTbeep();
	return FALSE;
    }

    /* the longestmatch array has the longest match so copy and print it */
    for (; (*cpos < (NSTRING - 1)) && (*cpos < longestlen); (*cpos)++)
    {
	name[*cpos] = longestmatch[*cpos];
	TTputc(name[*cpos]);
    }

    name[*cpos] = 0;


    /* MJH - I don't like this auto end for filenames since I may be
         * probing or assembling a new name - disabled
         */
#if AUTOENDFNM & NEWVAR
    /* 
     * If only one file matched then increment cpos to signal complete() 
     * that this was a complete match.  If a directory was matched then
     * last character will be the DIRSEPCHAR.  In this case we do NOT 
     * want to signal a complete match. 
     */
    if ((uconfig & UC_AUTOENDCOMP) &&
	(matches == 1) && (name[(*cpos) - 1] != DIRSEPCHAR))
    {
	(*cpos)++;
	retval = TRUE;
    }

#endif /* AUTOENDFNM */
    TTflush();

  do_clist:
    if (matches > 1)		/* multiple matches - show possibilities */
    {
	clist_file(name, cpos);
	update(TRUE);
    }

    return retval;
}

#endif /* !WINDOW_MSWIN */


/*	tgetc:	Get a key from the terminal driver, resolve any keyboard
 *	        macro action
 */

int tgetc()
{
    int c;		     /* fetched character */

    /* if we are playing a keyboard macro back, */
    if (kbdmode == PLAY)
    {

	/* if there is some left... */
	if (kbdptr < kbdend)
	    return ((int) *kbdptr++);

	/* at the end of last repitition? */
	if (--kbdrep < 1)
	{
	    kbdmode = STOP;
#if	VISMAC == 0
	    /* force a screen update after all is done */
	    update(FALSE);
#endif
	}
	else
	{

	    /* reset the macro to the begining for the next rep */
	    kbdptr = &kbdm[0];
	    return ((int) *kbdptr++);
	}
    }

    /* if no pending character */
    if (cpending == FALSE)
    {

	/* fetch a character from the terminal driver */
	c = TTgetc();

    }
    else
    {

	c = charpending;
	cpending = FALSE;
    }

    /* record it for $lastkey */
    lastkey = c;

    /* save it if we need to */
    if (kbdmode == RECORD)
    {
	*kbdptr++ = c;
	kbdend = kbdptr;

	/* don't overrun the buffer */
	if (kbdptr == &kbdm[NKBDM - 1])
	{
	    kbdmode = STOP;
	    TTbeep();
	}
    }

    /* and finally give the char back */
    return (c);
}

/*	getkey:	Get one keystroke. The legal prefixs here
			are the SPEC, MOUS and CTRL prefixes.
*/

int getkey()
{
    int c;		     /* next input character */
    int upper;		     /* upper byte of the extended sequence */

    /* get a keystroke */
    c = tgetc();

    /* if it exists, process an escape sequence */
    if (c == 0)
    {

	/* get the event type */
	upper = tgetc();

	/* mouse events need us to read in the row/col */
	if (upper & (MOUS >> 8))
	{
	    /* grab the x/y position of the mouse */
	    xpos = tgetc();
	    ypos = tgetc();
	}

	/* get the event code */
	c = tgetc();

	/* if it is a function key... map it */
	c = (upper << 8) | c;
    }

    /* yank out the control prefix */
    if (((c & 255) >= 0x00 && (c & 255) <= 0x1F) || (c & 255) == 0x7F)
	c = CTRL | (c ^ 0x40);

    /* return the character */
    return (c);
}

/*	GETCMD:	Get a command from the keyboard. Process all applicable
 *		prefix keys
 */
int getcmd()
{
    int c;		     /* fetched keystroke */
    KEYTAB *key;	     /* ptr to a key entry */

    /* get initial character */
    c = getkey();
    key = getbind(c);

    /* resolve META and CTLX prefixes */
    if (key)
    {
	if (key->k_ptr.fp == meta)
	{
	    c = getkey();
#if	SMOS
	    c = upperc(c & 255) | (c & ~255);	/* Force to upper */
#else
	    c = upperc(c) | (c & ~255);	/* Force to upper */
#endif
	    c |= META;
	}
	else if (key->k_ptr.fp == cex)
	{
	    c = getkey();
#if	SMOS
	    c = upperc(c & 255) | (c & ~255);	/* Force to upper */
#else
	    c = upperc(c) | (c & ~255);	/* Force to upper */
#endif
	    c |= CTLX;
	}
    }

    /* return it */
    return (c);
}

#if MSG_STACK
/* -------------------- MSG LINE STACKING RTNS ---------------------- */

#define MSGSTKSIZ	20   /* number of message lines to stack */

static char *msgstk[MSGSTKSIZ];
static int stktop = 0;	     /* TOS - next free */
static int stkcur = 0;	     /* element got */


/* stacks buf into message stack */
int stackmsg(buf, len)
     char *buf;		     /* line to stack */
     int len;		     /* line length */
{
    char *s;
    int i;

    if (msgstk[stktop] != NULL)
    {
	free(msgstk[stktop]);
	msgstk[stktop] = NULL;
    }

    i = stktop - 1;		/* line on top of stack */
    if (i < 0)
	i = MSGSTKSIZ - 1;
    if (msgstk[i] != NULL && strcmp(msgstk[i], buf) == 0)
	return TRUE;		/* don't restack same value */

    if ((s = (char *) malloc(len + 1)) == NULL)
    {
	mlwrite("msgstack malloc fail");
	return FALSE;
    }
    strcpy(s, buf);
    msgstk[stktop++] = s;
    stktop %= MSGSTKSIZ;
    stkcur = stktop;
    return TRUE;
}

/*
 *  gets next msg in stack in dir given into buf
 *  returns TRUE if unstacked message, FALSE otherwise
 */
int unstackmsg(buf, dir)
     char *buf;
     int dir;		     /* 1 if want next, -1 if want previous */
{


    if ((dir == 1 && stkcur == stktop) ||
	(dir == -1 && stkcur == (stktop + 1) % MSGSTKSIZ))
	return FALSE;		/* at edge */

    stkcur += dir;
    stkcur %= MSGSTKSIZ;
    if (stkcur < 0)
	stkcur = MSGSTKSIZ - 1;
    if (dir == 1 && stkcur == stktop)
	return FALSE;
    while (msgstk[stkcur] == NULL)
    {
	stkcur += dir;
	stkcur %= MSGSTKSIZ;
	if (stkcur < 0)
	    stkcur = MSGSTKSIZ - 1;

	if (dir == -1 && stkcur == (stktop + 1) % MSGSTKSIZ)
	{
	    if (msgstk[stkcur] == NULL)
		stkcur = 0;
	    return FALSE;
	}
	if (dir == 1 && stkcur == stktop)
	{
	    return FALSE;
	}
    }
    strcpy(buf, msgstk[stkcur]);
    return TRUE;
}

#endif /* MSG_STACK */


/* -------------------- Messageline Display fns ----------------------- */

/* calculate offset on screen of posn cpos in buf */
static int mlscnofst(buf, cpos, adjust)
     char *buf;
     int cpos,
         adjust;
{
    register int col = adjust;
    register int i;

    for (i = 0; i < cpos; ++i)
    {
	if (buf[i] == '\n')	/* "<NL>" */
	    col += 3;
	else if (buf[i] < ' ')	/* "^?" */
	    ++col;
	++col;			/* letter */
    }
    return col;
}

/*
 * move to a specified col on the message line
 */
static void mlmoveto(col)
{
    movecursor(term.t_nrow, col);
    TTflush();
}


/* put a message line character to  the display with ctrl char expansion */
static void mldsp1ch(c)
     int c;
{
    if ((c < ' ') && (c != '\n'))
    {
	ostring("^");
	++ttcol;
	c ^= 0x40;
    }
    if (c != '\n')
    {
	if (disinp)
	    TTputc(c);
    }
    else
    {				/* put out <NL> for <ret> */
	ostring("<NL>");
	ttcol += 3;
    }
    ++ttcol;
}

/*
 * redisplay remainder of msg line buffer from ofst cpos
 */
static void mldspln(cpos, len, buf, csr2eoln)
     int cpos,
         len;
     char *buf;
     int csr2eoln;	     /* TRUE if want display csr left at eoln */
{
    register int i;
    register int ottcol;

    ottcol = ttcol;
    for (i = cpos; i < len; ++i)
	mldsp1ch((int) *(buf + i));
    if (!csr2eoln)
    {
	i = ttcol;
	ttcol = 0;
	mleeol(i);
	mlmoveto(ottcol);
    }
    TTflush();
}


/*
 * delete char under cursor, close up line buffer and display
 */
static void mldel(cpos, len, buf)
     int cpos,
         len;
     char *buf;
{
    register int i;

    for (i = cpos; i < len; ++i)
	*(buf + i) = *(buf + i + 1);
    mldspln(cpos, len - 1, buf, FALSE);
}

/*
 * Put a character to the display and the line buffer
 * Update remainder of display line if necessary
 * return pos in line buffer
 */
static int mlputch(cpos, c, buf, len, ins)
     int cpos,
         c;
     char *buf;
     int len;
     int ins;		     /* TRUE if inserting, FALSE if Displaying */
{
    register int i;

    if (ins && cpos < len)	/* move everthing up */
    {
	for (i = len; i >= cpos; --i)
	    *(buf + i + 1) = *(buf + i);
	++len;
    }
    buf[cpos++] = c;		/* put char in buffer */
    mldsp1ch(c);		/* display it */

    if (cpos < len)
	mldspln(cpos, len, buf, FALSE);	/* upd rest of display */

    TTflush();
    return cpos;
}

/*
 * Append a string onto the end of the buffer and update display leaving csr
 * at end of string
 * NOTE : maxpos must = cpos  when enter this rtn
 * return end of buffer ( new maxpos, cpos )
 */
static int mlapnds(buf, fm, cpos, nbuf)
     char *buf,
         *fm;
     int cpos,
         nbuf;
{
    int i,
        maxpos;

    i = strlen(fm);
    if (cpos + i < nbuf - 1)
    {
	buf[cpos] = '\0';
	strcat(buf, fm);	/* add string to bfr */
	maxpos = cpos + i;
    }
    else
    {
	i = nbuf - cpos - 12;
	buf[cpos] = '\0';
	strncat(buf, fm, i);	/* add string to bfr */
	maxpos = cpos + i;
	buf[maxpos] = '\0';
    }
    mldspln(cpos, maxpos, buf, TRUE);	/* update display */
    return maxpos;
}


#if INTERNAL_METACHARS
/*
 * get substitution string corresponding to character entered
 * returns TRUE if valid subs char and fills string ptr, FALSE otherwise
 *   %% - single %
 *   %b - curr buffer name
 *   %f - curr buffer filename
 *   %i - '/usr/include'
 *   %s - curr search string
 *   %r - curr replacement string
 *++ these should be user extendable from Macro lang.
 */
static int get_subs_val(c, sp)
     char c;
     char **sp;
{
    c = ectoc(getcmd());
    switch (c)
    {
    case '%':			/* %% -> '%' */
	*sp = "%";
	break;
    case 'b':			/* buffer name */
	*sp = curbp->b_bname;
	break;
    case 'f':			/* buffer filename */
	*sp = curbp->b_fname;
	break;
    case 'i':			/* /usr/include */
	*sp = "/usr/include";
	break;
    case 'r':			/* replacement string */
	*sp = (char *) rpat;
	break;
    case 's':			/* search string */
	*sp = (char *) pat;
	break;
    default:
	TTbeep();
	TTflush();
	return FALSE;
	/* NOTREACHED */ break;
    }
    return TRUE;
}

#endif /* INTERNAL_METACHARS */

#if ENVFUNC
/*
 * Handle ~ and ~user expansion
 * Display only work at end of minibfr
 * return TRUE if successful, FALSE if abort for any reason, FAIL otherwise
 */
int do_tilde(buf, spos, nbuf, pcpos)
     char *buf;
     int spos;
     int nbuf;
     int *pcpos;
{
    char *home_ptr;	     /* pointer to home directory string */
    char *ptr;		     /* string pointer */
    char user_name[NSTRING]; /* user name for directory */
    int cpos = *pcpos;

    if ((home_ptr = getenv("HOME")) == (char *) NULL)
	return -1;

    /* save the user name! */
    buf[cpos] = '\0';
    strcpy(user_name, &buf[spos + 1]);

    /* erase the chars on-screen */
    buf[spos] = '\0';
    ttcol -= cpos - spos;
    mldspln(spos, strlen(buf), buf, FALSE);
    cpos = spos;

#if USG | AUX | BSD | V7 | SUN | HPUX8 | HPUX9
    /* lookup someone else's home directory! */
    if (user_name[0] != '\0')
    {
	struct passwd *pwd;  /* password structure */

	pwd = getpwnam(user_name);
	if (pwd != (struct passwd *) NULL)
	{
	    ptr = pwd->pw_dir;
	    cpos = mlapnds(buf, ptr, cpos, nbuf);
	}
    }
#endif

    if (cpos == spos)
    {
	/* output the home directory */
	cpos = mlapnds(buf, home_ptr, cpos, nbuf);

	/* is this someone else's home directory */
	if (user_name[0] != '\0')
	{
	    spos = cpos;
	    /* backup to the last directory sep */
	    while ((cpos > 0) && (buf[cpos - 1] != DIRSEPCHAR))
		--cpos;

	    buf[cpos] = '\0';
	    ttcol -= spos - cpos;
	    mldspln(cpos, strlen(buf), buf, FALSE);

	    /* and add the user's name */
	    cpos = mlapnds(buf, user_name, cpos, nbuf);
	}
    }
#if 0
    /* and the last directory seperator */
    if (cpos > 0 && buf[cpos - 1] != DIRSEPCHAR)
	cpos = mlputch(cpos, DIRSEPCHAR, buf, cpos, FALSE);
#endif

    *pcpos = cpos;
    return TRUE;
}

/*
 * Handle expansion of environ vars ( start with leading $ )
 * return TRUE if successful, FALSE if abort for any reason, FAIL otherwise
 */
int do_envvar(buf, spos, nbuf, pcpos)
     char *buf;
     int spos;
     int nbuf;
     int *pcpos;
{
    int cpos = *pcpos;
    char env_name[NSTRING];  /* user name for directory */
    char *envptr;

    if (cpos <= spos)
	return FALSE;

    /* expand an environment variable reference */
    /* save the variable name! */
    buf[cpos] = '\0';
    strcpy(env_name, &buf[spos + 1]);
#if	MSDOS | OS2 | VMS
    mkupper(user_name);
#endif

    buf[spos] = '\0';
    ttcol -= cpos - spos;
    mldspln(spos, strlen(buf), buf, FALSE);
    cpos = spos;

    envptr = getenv(env_name);
    if (envptr == (char *) NULL)
	return FALSE;

    cpos = mlapnds(buf, envptr, cpos, nbuf);

    *pcpos = cpos;
    return TRUE;
}

/*
 * Return TRUE and ofset of word start if word ending at
 * cpos starts with interesting char.
 * else returns FALSE
 */
int chk_special_startwd(buf, cpos, spos)
     char *buf;
     register int cpos;
     int *spos;
{
    for (; cpos >= 0 && buf[cpos] != ' '; --cpos)
    {
	if ((buf[cpos] == '~' || buf[cpos] == '$') &&
	    (cpos == 0 || buf[cpos - 1] == ' '))
	{
	    *spos = cpos;
	    return TRUE;
	}
    }
    return FALSE;
}

#endif /* ENVFUNC */


/*
	A more generalized prompt/reply function allowing the caller
	to specify the proper terminator. If the terminator is not
	a return ('\n') it will echo as "<NL>"

	Extensively modified from default for msgline editing, stacking
	and % expansion
*/
int getstring(prompt, buf, nbuf, eolchar, cmp_type)
     char *prompt;
     char *buf;
     int eolchar;
     int cmp_type;	     /* Name Completion Type */
{
    register int cpos;	     /* current character position in string */
    register int ec;	     /* extended current input character */
    register int c;
    register int quotef;     /* are we quoting the next char? */
    register int maxpos;
    char pbuf[NSTRING];
    int plen;
    int (*efn) ();	     /* ptr to function bound to key */
    KEYTAB *ktp;
    int ofst;
    int s_wd = -1;	     /* start of word with interesting start char */

    if (eolchar == CTRL | '@')	/* empty char equiv */
	eolchar = (CTRL | 'M');

    cpos = maxpos = 0;
    quotef = FALSE;

    /* prompt the user for the input string */
    if (discmd)
    {
	if (prompt)
	    mlwrite(prompt);
    }
    else
	mlmoveto(0);

    strcpy(pbuf, prompt);
    plen = strlen(pbuf);

    for (;;)
    {
	/* get a character from the user */
	if (quotef == FALSE)
	{
	    ec = getcmd();
	}
	else
	{
	    ec = getkey();
	    c = ectoc(ec);
	    quotef = FALSE;
	    if (cpos < nbuf - 1)
	    {
		cpos = mlputch(cpos, c, buf, maxpos, TRUE);
		++maxpos;
	    }
	    continue;
	}
	/* if the keystroke is a bound function...get its binding */
	efn = NULL;
	ktp = getbind(ec);

	/* cannot go directly to self_insert code from here if
         * ktp is NULL ( assumes self-insert ) since
         * we overload some of the characters if they are at the end
         * of the minibfr - '%' for e.g.
         */
	if (ktp != NULL)
	    if (ktp->k_type == BINDFNC)
		efn = ktp->k_ptr.fp;
	    else
		continue;


/*
 * Minibuffer binding table - this should be made table driven same as
 * bindings in the main buffers
 */
	/* if they hit the line terminate, wrap it up */
	if (efn == newline || ec == eolchar)
	{
	    buf[maxpos] = 0;

	    /* clear the message line */
	    mlwrite("");
	    TTflush();

	    /* if we default the buffer, return FALSE */
	    if (buf[0] == 0)
		return (FALSE);
#if MSG_STACK
	    stackmsg(buf, maxpos);
	    stkcur = stktop;
#endif

	    return (TRUE);
	}

	/* change from command form back to character form */
	c = ectoc(ec);
	if (ec & CTRL)
	    c = cntl(c);

	/* if it is from the mouse, or is a function key, blow it off */
	if ((quotef == FALSE) && ((ec & MOUS) /* || (ec & SPEC) */ ))
	    continue;



	if (efn == ctrlg)
	{			/* Abort the input? */
#if MSG_STACK
	    stkcur = stktop;
#endif
	    ctrlg(FALSE, 0);
	    TTflush();
	    return (ABORT);
	}
	else if (efn == backdel)/* erase */
	{
	    if (cpos > 0)
	    {
		mlmoveto(mlscnofst(buf, --cpos, plen));
		mldel(cpos, maxpos, buf);
		--maxpos;
	    }
	}
	else if (efn == forwdel)/* forward del */
	{
	    if (cpos < maxpos)
	    {
		mldel(cpos, maxpos, buf);
		--maxpos;
	    }
	}
	else if (efn == delfword)	/* word kill */
	{
	    char c = buf[cpos];

	    while (cpos < maxpos &&	/* note shitty defn of a word */
		   ((c == ' ' && buf[cpos] == c) ||
		    (c != ' ' && buf[cpos] != ' ')))
	    {
		mldel(cpos, maxpos, buf);
		--maxpos;
	    }
	}
	else if (c == cntl('U') ||	/* fake line kill */
		 efn == killtext/* kill to eoln */
#if LINE_FNS
		 || efn == kill_line	/* kill entire line */
#endif
	    )
	{
	    mlwrite(pbuf);	/* clears data on line */
	    TTflush();
	    /* Clearing empty line ? - kill buffer and
             * return TRUE. Allows specifying empty string
             * for text replacement string instead of
             * defaulting to last value used ( ^K^K ).
             */
	    if (cpos == 0 && buf[0] == '\0')
	    {
		/* clear the message line and leave */
		buf[0] = 0;
		mlwrite("");
		TTflush();
		return (TRUE);
	    }
	    if (efn == killtext)
	    {
		maxpos = cpos;
		mldspln(0, maxpos, buf, TRUE);
	    }
	    else
		maxpos = cpos = 0;
	    buf[cpos] = '\0';
	}
	else if (efn == backchar)
	{
	    if (cpos > 0)
	    {
		--cpos;
		mlmoveto(mlscnofst(buf, cpos, plen));
	    }
	}
	else if (efn == forwchar)
	{
	    if (cpos < maxpos)
	    {
		++cpos;
		mlmoveto(mlscnofst(buf, cpos, plen));
	    }
	}
	else if (efn == gotobol)/* Soln */
	{
	    cpos = 0;
	    mlmoveto(mlscnofst(buf, cpos, plen));
	}
	else if (efn == gotoeol)/* Eoln */
	{
	    cpos = maxpos;
	    mlmoveto(mlscnofst(buf, cpos, plen));
	}
	else if (efn == backword)	/* word back */
	{
	    if (cpos > 0 && buf[cpos - 1] == ' ')
		--cpos;
	    while (cpos > 0)
	    {
		--cpos;
		if (cpos > 0 && buf[cpos] != ' ' && buf[cpos - 1] == ' ')
		    break;
	    }
	    mlmoveto(mlscnofst(buf, cpos, plen));
	}
	else if (efn == forwword)	/* Word fwd */
	{
	    if (cpos < maxpos && buf[cpos - 1] == ' ')
		++cpos;
	    while (cpos < maxpos)
	    {
		cpos++;
		if (buf[cpos] != ' ' && buf[cpos - 1] == ' ')
		    break;
	    }
	    mlmoveto(mlscnofst(buf, cpos, plen));
	}
	else if (efn == refresh)/* repaint msg line */
	{
	    mlwrite(prompt);
	    mldspln(0, maxpos, buf, TRUE);
	}
#if MSG_STACK
	else if (efn == backline || efn == forwline)
	{			/* Previous line       Next Line */
	    c = (efn == backline) ? -1 : 1;
	    if (unstackmsg(buf, c) == TRUE)
	    {
		mlwrite(pbuf);
		cpos = 0;
		maxpos = strlen(buf);
		mldspln(cpos, maxpos, buf, TRUE);
		cpos = maxpos;
	    }
	    else
	    {
		TTbeep();
		TTflush();
	    }
	}
#endif /* MSG_STACK */
	else if (efn == quote)
	{
	    quotef = TRUE;
	}
	else if (cpos == maxpos)
	{		/* chars with special meaning only at end of minibf */

	    if (efn == tab ||
		(c == ' ' && cmp_type != CMP_DEFAULT) || (ec == sterm)
		)
		c = TMP_COMPCHAR;

	    switch (c)
	    {
	    case TMP_COMPCHAR:	/* complete */
		{
		    int lcpos = cpos;

		    buf[lcpos] = '\0';
		    switch (cmp_type)
		    {
		    case CMP_BUFFER:
			comp_buffer(buf, &lcpos);
			break;
		    case CMP_COMMAND:
			if (comp_command(buf, &lcpos))
			{	/* successfully matched */
			    buf[lcpos] = '\0';
#if AUTOENDFNM & NEWVAR
			    if ((uconfig & UC_AUTOENDCOMP))
				return TRUE;
#endif /* AUTOENDFNM & NEWVAR */
			}
			else
			    TTbeep();	/* no match or multiple matches */
			break;
#if	!WINDOW_MSWIN
		    case CMP_DEFAULT:
		    case CMP_FILENAME:

			/*
                         * Find start of name to match on since may have buf
                         * with embedded spaces in it
                         */
			for (ofst = cpos; ofst > 0; --ofst)
			{
			    if (buf[ofst - 1] == ' ')
				break;
			}

			lcpos -= ofst;
			if (comp_file(buf + ofst, &lcpos))
			{
#if AUTOENDFNM & NEWVAR
			    if ((uconfig & UC_AUTOENDCOMP))
				return TRUE;
#endif /* AUTOENDFNM & NEWVAR */
			}
			lcpos += ofst;
			break;
#endif /* !WINDOW_MSWIN */
		    }
		    buf[lcpos] = '\0';
		    maxpos = cpos = lcpos;	/* force redisplay of msgln */
		    mlwrite(prompt);
		    mldspln(0, maxpos, buf, TRUE);
		}
		break;

	    case '?':		/* Show possible completion list */
		{
		    int lcpos = cpos;

		    switch (cmp_type)
		    {
		    case CMP_BUFFER:
			clist_buffer(buf, &lcpos);
			break;
		    case CMP_COMMAND:
			clist_command(buf, &lcpos);
			break;
#if	!WINDOW_MSWIN
		    case CMP_FILENAME:
			clist_file(buf, &lcpos);
			break;
#endif
		    }
		}
		update(TRUE);
		mlwrite(prompt);
		mldspln(0, maxpos, buf, TRUE);
		break;

#if INTERNAL_METACHARS
	    case '%':		/* expansion of internal %c to string */
		{
		    char *s = NULL;

		    if (get_subs_val(c, &s))
			maxpos = cpos = mlapnds(buf, s, maxpos, nbuf);
		}
		break;
#endif /* INTERNAL_METACHARS */

#if ENVFUNC
	    case DIRSEPCHAR:
	    case ' ':
		if (chk_special_startwd(buf, cpos, &s_wd))
		{
		    int lcpos = cpos;
		    int done_ok = FALSE;

		    if (buf[s_wd] == '~')
			do_tilde(buf, s_wd, nbuf, &lcpos);
		    else if (buf[s_wd] == '$')
			do_envvar(buf, s_wd, nbuf, &lcpos);

		    maxpos = cpos = lcpos;
		    s_wd = -1;
		}
#endif /* ENVFUNC */
		/* fall through */

	    default:
		goto accept_char;
		break;		/* NOTREACHED */
	    }			/* switch */
	}			/* chars at end of minibuffer */
	else
	{
	  accept_char:
	    quotef = FALSE;
	    if (cpos < nbuf - 1)
	    {
		cpos = mlputch(cpos, c, buf, maxpos, TRUE);
		++maxpos;
	    }
	}
    }				/* for */
}


ostring(s)	     /* output a string of output characters */
     char *s;		     /* string to output */
{
    if (discmd)
	while (*s)
	    mlout(*s++);
}


/*
 * echostring -- Use echochar() to put out a string.  Checks for NULL.
 */
int echostring(str, tcol, uptocol)
     char *str;		     /* characters to be echoed */
     int tcol;		     /* column to be echoed in */
     int uptocol;	     /* last column to be echoed in */
{
    if (str != NULL)
    {
	while (*str)
	{
	    tcol = echochar(*str++, tcol);
	    if (tcol >= uptocol)
	    {
		mlout('$');
		tcol++;
		break;
	    }
	}
    }
    return (tcol);
}

/*
 * Routine to echo i-search and message-prompting characters.
 */
int echochar(c, col)
     unsigned char c;	     /* character to be echoed */
     int col;		     /* column to be echoed in */
{
    movecursor(term.t_nrow, col);	/* Position the cursor	*/
    if (c == '\r')
    {				/* Newline character?	*/
	mlout('<');
	mlout('N');
	mlout('L');
	mlout('>');
	col += 3;
    }
#if 0
    else if (c == '\t')
    {				/* Tab character?	*/
	mlout('<');
	mlout('T');
	mlout('A');
	mlout('B');
	mlout('>');
	col += 4;
    }
#endif
    else if ((c < ' ') || (c == 0x7F))
    {				/* Vanilla control char and Rubout:   */
	mlout('^');		/* Yes, output prefix		*/
	mlout(c ^ 0x40);	/* Make it "^X"			*/
	col++;			/* Count this char		*/
    }
    else
	mlout(c);		/* Otherwise, output raw char	*/
    TTflush();			/* Flush the output		*/
    return (++col);		/* return the new column number */
}
