/*
 *	UNDO for MicroEMACS 3.12
 *
 *	(C)Copyright 1994 by Jim Kissel
 *
 *	MicroEMACS 3.12 can be copied and distributed freely for any
 *	non-commercial purposes. MicroEMACS 3.12 can only be incorporated
 *	into commercial software with the permission of the current author.
 *
 */

#include	<stdio.h>
#include	"estruct.h"
#include	"eproto.h"
#include	"edef.h"
#include	"elang.h"

#define	BSIZE(a)	(a + NBLOCK - 1) & (~(NBLOCK - 1))
#define	Char_insert(a)	(a == '\r' ? lnewline() : linsert(1, a))

#if UNDOCMD
/*
 *	undo: entry point into undoing things
 */
PASCAL NEAR undo(f, n)

int f,n;	/* prefix flag and argument */

{
	undo_in_progress = TRUE;

	if( curbp->undo_current->undo_f & UNDO_REPLACE )
	{
		curbp->undo_current->undo_f &= (~UNDO_REPLACE);
		do
		{
			curbp->undo_current->n_func(curbp->undo_current->undo_f,
										curbp->undo_current->undo_n);
			curbp->undo_current = curbp->undo_current->next_undo;
		} while( !(curbp->undo_current->undo_f & UNDO_REPLACE) );
	}
	else
	{
		if( curbp->undo_current->undo_f & UNDO_QUERY_REPLACE )
		{
			curbp->undo_current->undo_f &= (~UNDO_QUERY_REPLACE);
			if( n == 1 )
				n = 3;	/* the movement, the delete, and the replace */
		}
		while(n--)
		{
			if( curbp->undo_tail == curbp->undo_current)
			{
				undo_finished(TRUE, 0);
			}
			else
			{
				curbp->undo_current->n_func(curbp->undo_current->undo_f,
											curbp->undo_current->undo_n);
				curbp->undo_current = curbp->undo_current->next_undo;
			}
		}
	}

	/* if we have undone all real changes */
	if( curbp->b_modification_cnt == 0 )
		unmark(f, n);
	undo_in_progress = FALSE;
	return(TRUE);
}

/*
 *	nothing left to undo
 */
 
PASCAL NEAR undo_finished(f, n)

int f,n;	/* prefix flag and argument */

{
	mlwrite(TEXT350);	/* Nothing to undo */
	return(TRUE);
}

/*
 *	save the dot for use later
 */

PASCAL NEAR undo_save_dot(f, n)

int f,n;	/* prefix flag and argument */
{
	struct	UNDO *undoptr;

	if( (undoptr = (UNDO *)malloc(sizeof(UNDO))) == NULL )
		return(NULL);
	undoptr->next_undo = curbp->undo_current;	/* point at last cmd */
	curbp->undo_current = undoptr;
	undoptr->b_dotp = curwp->w_dotp;
	undoptr->b_doto = curwp->w_doto;
	undoptr->undo_n = 0;
	undoptr->undo_f = f;
	undoptr->n_func = undo_gotomark;
}
/*
 * goto the mark contained in the undo struct passed
 */
 
int PASCAL NEAR undo_gotomark(f, n)

int f,n;	/* prefix flag and argument */

{
	curwp->w_dotp  = curwp->w_bufp->undo_current->b_dotp;
	curwp->w_doto  = curwp->w_bufp->undo_current->b_doto;
	curwp->w_flag |= WFMOVE;
	return(TRUE);
}

/*
 *	save a function for later use
 *	if f is TRUE && last cmd is the same as the current command
 *  just increment the undo count
 *		otherwise
 *	allocate a new undo structure
 */

PASCAL NEAR undo_save_fn(f, n, save_fn )

int f, n, (PASCAL NEAR *save_fn)();

{
	struct	UNDO *undoptr;

	if( (curbp->undo_current->n_func != save_fn) ||
		(curbp->undo_current->n_func == save_fn && f == FALSE) )
	{
		if( (undoptr = (UNDO *)malloc(sizeof(UNDO))) == NULL )
		{
			/* evenutally we need to realloc some of the undo space
			 * if we get here.
			 */
			return(NULL);
		}
		undoptr->next_undo = curbp->undo_current;	/* point at last cmd */
		curbp->undo_current = undoptr;
		undoptr->undo_n = n;
		undoptr->undo_f = f;
		undoptr->n_func = save_fn;
		
		/* the following not always required but....... */
        undoptr->b_dotp = curwp->w_dotp;
        undoptr->b_doto = curwp->w_doto;

        /* change to buffer contents. increment the modification cnt */
		if ( save_fn == undo_linsert || save_fn == undo_newline ||
				save_fn == undo_ifile )
			curbp->b_modification_cnt++;
		return(f);
	}
	else
	{
		if(save_fn == undo_linsert )
		{
			curbp->undo_current->b_dotp = curwp->w_dotp;
			return(TRUE);
		}
	}
	/* not always used though */
	curbp->undo_current->undo_n += n;
	curbp->undo_current->undo_f = f;
}

/*
 * undo a character insert
 */
 
PASCAL NEAR undo_linsert(f, n)

int f, n;

{
	LINE	*lp;
	short	ldot;

	/* save mark[0] if set */
	lp = curwp->w_markp[0];
	ldot = curwp->w_marko[0];

	/* set mark[0] to stored undo mark */
	curwp->w_markp[0] = curwp->w_bufp->undo_current->b_dotp;
	curwp->w_marko[0] = curwp->w_bufp->undo_current->b_doto;

	/* del the region to get rid of inserted text */
	killregion(f, n);

	curwp->w_markp[0] = lp;
	curwp->w_marko[0] = ldot;
	curbp->b_modification_cnt--;
}

/*
 * undo a newline
 */

PASCAL NEAR undo_newline(f, n)

int f, n;

{
	backdel(f, n);
	curbp->b_modification_cnt--;
}

/*
 *	check through the undo list for structs which may have had their
 *	LINE reallocated and update these with the new pointer
 */
 
PASCAL NEAR fix_undo_struct(oldline, newline)

LINE *oldline, *newline;

{
	UNDO	*start_ptr, *stop_ptr;
	
	start_ptr = curbp->undo_current;
	stop_ptr = curbp->undo_tail;
	while(start_ptr != stop_ptr)
	{
		if( start_ptr->b_dotp == oldline )
		{
			start_ptr->b_dotp = newline;
		}
		start_ptr = start_ptr->next_undo;
	}
}

/*
 * save deleted chars for later use
 */
PASCAL NEAR undo_save_ldelete(f, n, cp1, cp2)

int f, n;
char *cp1, *cp2;
{
	struct	UNDO *undoptr;
	LINE *lp1, *lp2, *lp3;
	char *cp3, *cp4;

	undoptr = curbp->undo_current;
	lp1 = undoptr->b_dotp;
	
	if( undoptr->n_func == undo_ldelete && f == undoptr->undo_f ) 
	{
		if( lp1->l_used + (cp2 - cp1) > lp1->l_size)
		{
	        if( (lp2 = lalloc(BSIZE(lp1->l_size + (cp2 - cp1)))) == NULL )
			{
				/* evenutally we need to realloc some of the undo space
				 * if we get here.
				 */
				return(NULL);
			}
			undoptr->b_dotp = lp2;
			/* the orginal saved text over */
			strncpy(lp2->l_text, lp1->l_text, lp1->l_used);
			lp2->l_used = lp1->l_used;
			free( (char *) lp1);
			lp1 = lp2;
		}
		/* then copy the new text over */
        strncpy(&lp1->l_text[lp1->l_used] , cp1, cp2 - cp1);
		lp1->l_used += cp2 - cp1;
	}
	else
	{
		if( (undoptr = (UNDO *)malloc(sizeof(UNDO))) == NULL )
		{
			/* evenutally we need to realloc some of the undo space
			 * if we get here.
			 */
			return(NULL);
		}
		undoptr->next_undo = curbp->undo_current;	/* point at last cmd */
		curbp->undo_current = undoptr;
		undoptr->undo_n = n;
		undoptr->undo_f = f;
		undoptr->n_func = undo_ldelete;

		/* lalloc a line to store the deleted chars */
        if( (undoptr->b_dotp = lalloc(BSIZE(n))) == NULL )
		{
			/* evenutally we need to realloc some of the undo space
			 * if we get here.
			 */
			return(NULL);
		}
        strncpy( undoptr->b_dotp->l_text, cp1, cp2 - cp1);
		undoptr->b_dotp->l_used = cp2 - cp1;
		curbp->b_modification_cnt++;
	}
}

/*
 * undo delete character(s)
 */
 
PASCAL NEAR undo_ldelete(f, n)

int f, n;
{
	struct	UNDO *undoptr;
	LINE *lp1, *lp2;
	int	i, doto;

	/* save current dot: thanks to yank() */
	lp2 = lback(curwp->w_dotp);
	doto = curwp->w_doto;

	undoptr = curbp->undo_current;
	lp1 = undoptr->b_dotp;
	if( undoptr->undo_f == TRUE )	/* a backdel */
	{
		for(i = lp1->l_used; i; i--)
			Char_insert(lp1->l_text[i -1]);
	}
	else
	{
		for(i = 0; i < lp1->l_used; i++)
			Char_insert(lp1->l_text[i]);
	}

	curbp->b_modification_cnt--;

	/* restore the dot */
	curwp->w_dotp = lforw(lp2);
	curwp->w_doto = doto;
	if( undoptr->undo_f == TRUE )	/* a backdel */
		forwchar(TRUE, lp1->l_used);
#ifdef notdef
	if( n < 0 )	/* delete word backward would leave an incorrect dot */
		curwp->w_doto += abs(n);
#endif
	curwp->w_flag |= WFMOVE;
}

PASCAL NEAR undebug(f, n)

int f, n;
{
	char foo[100];

	sprintf(foo,"line id == %ld", curwp->w_dotp);
	mlwrite(foo);
}

PASCAL NEAR undump(f,n)

int f, n;

{
	char	foo[100];
	UNDO	*start_ptr, *stop_ptr;
	
	fprintf(stderr,"\n\n");
	start_ptr = curbp->undo_current;
	stop_ptr = curbp->undo_tail;
	while(start_ptr != stop_ptr)
	{
		sprintf(foo,"line id == %ld dot == %d f == %d n == %d func->",
					start_ptr->b_dotp,
					start_ptr->b_doto,
					start_ptr->undo_f,
					start_ptr->undo_n);
		if(start_ptr->n_func == undo_newline )
			strcat(foo, "undo_newline");
		if(start_ptr->n_func == undo_linsert )
			strcat(foo, "undo_linsert");
		if(start_ptr->n_func == undo_gotomark )
			strcat(foo, "undo_gotomark");
		if(start_ptr->n_func == undo_finished )
			strcat(foo, "undo_finished");
		if(start_ptr->n_func == undo_save_dot )
			strcat(foo, "undo_savedot");
		if(start_ptr->n_func == forwchar )
			strcat(foo, "forwchar");
		if(start_ptr->n_func == backchar )
			strcat(foo, "backchar");
		if(start_ptr->n_func == backline )
			strcat(foo, "backline");
		if(start_ptr->n_func == forwline )
			strcat(foo, "forwline");
		if(start_ptr->n_func == remmark )
			strcat(foo, "remmark");
		if(start_ptr->n_func == setmark )
			strcat(foo, "setmark");
		if(start_ptr->n_func == detab )
			strcat(foo, "detab");
		if(start_ptr->n_func == entab )
			strcat(foo, "entab");
		strcat(foo,"\n");
		fprintf(stderr,"%s",foo);

		start_ptr = start_ptr->next_undo;
	}
}

/*
 * when we saved the fn undo_file, curbp->b_doto was pointing at the
 * line before where the file was inserted so...........
 * fake out undo_linsert and then setup the correct dot
 *
 * no need to decrement the modification count as undo_linsert() does it
 */
PASCAL NEAR undo_ifile(f,n)

int f, n;

{
	LINE *curlp, *tmplp;
	int	curdot, tmpdot;

	curdot  = curwp->w_bufp->undo_current->b_doto;
	curwp->w_bufp->undo_current->b_doto = 0;
	curwp->w_bufp->undo_current->b_dotp =
				lforw(curwp->w_bufp->undo_current->b_dotp);
	undo_linsert(f,n);
 	curwp->w_doto  = curdot;
}

#endif /* UNDOCMD */
